import typing
from pathlib import PurePath
from typing import Optional, List, Union, Tuple, Dict

from .destinations.factory import create_destination
from .inspect import Inspect
from .iterator import Iterator
from .loop import Loop
from .sections import Section, OuterSection
from .store import Store
from .store.indicators import Indicator
from .writers import Writer
from .writers.screen import ScreenWriter
from ..util.colors import StyleCode
from ...logger import Text


class Logger:
    """
    This handles the interactions among sections, loop and store
    """

    def __init__(self):
        self.__store = None
        self.__writers: List[Writer] = []

        self.__loop: Optional[Loop] = None
        self.__sections: List[Section] = []

        self.__indicators_print = []

        self.__screen_writer = ScreenWriter()
        self.__writers.append(self.__screen_writer)

        self.__start_global_step: Optional[int] = None
        self.__global_step: Optional[int] = None
        self.__last_global_step: Optional[int] = None

        self.__destination = create_destination()
        self.__inspect = Inspect(self)

    @property
    def _store(self):
        if self.__store is None:
            self.__store = Store()

        return self.__store

    @property
    def global_step(self) -> int:
        if self.__global_step is not None:
            return self.__global_step

        global_step = 0
        if self.__start_global_step is not None:
            global_step = self.__start_global_step

        if self.__is_looping():
            return global_step + self.__loop.counter

        if self.__last_global_step is not None:
            return self.__last_global_step

        return global_step

    def add_writer(self, writer: Writer):
        self.__writers.append(writer)

    def reset_writers(self):
        self.__writers = []
        self.__writers.append(self.__screen_writer)

    def log(self, parts: List[Union[str, Tuple[str, Optional[StyleCode]]]], *,
            is_new_line=True):
        self.__destination.log(parts, is_new_line=is_new_line)

    def reset_store(self):
        self.__store = Store()

    def add_indicator(self, indicator: Indicator):
        self._store.add_indicator(indicator)

    def save_indicators(self, file: PurePath):
        self._store.save_indicators(self.__writers, file)

    def store(self, key: str, value: any):
        self._store.store(key, value)

    def store_namespace(self, name: str):
        return self._store.create_namespace(name)

    def set_global_step(self, global_step: Optional[int]):
        self.__global_step = global_step

    def set_start_global_step(self, global_step: Optional[int]):
        self.__start_global_step = global_step

    def add_global_step(self, increment_global_step: int = 1):
        if self.__global_step is None:
            if self.__start_global_step is not None:
                self.__global_step = self.__start_global_step
            else:
                self.__global_step = 0

        self.__global_step += increment_global_step

    def __is_looping(self):
        if self.__loop is not None and self.__loop.is_started:
            return True
        else:
            return False

    def write_h_parameters(self, hparams: Dict[str, any]):
        for w in self.__writers:
            w.write_h_parameters(hparams)

    def write(self):
        global_step = self.global_step

        self._store.save_indicators(self.__writers)

        for w in self.__writers:
            if w != self.__screen_writer:
                self._store.write(w, global_step)
        self.__indicators_print = self._store.write(self.__screen_writer, global_step)
        self._store.clear()

        parts = [(f"{self.global_step :8,}:  ", Text.highlight)]
        if self.__is_looping():
            self.__log_looping_line()
        else:
            parts += self.__indicators_print
            self.log(parts, is_new_line=False)

    def iterate(self, name, iterable: Union[typing.Iterable, typing.Sized, int],
                total_steps: Optional[int], *,
                is_silent: bool,
                is_children_silent: bool,
                is_timed: bool):
        return Iterator(logger=self,
                        name=name,
                        iterable=iterable,
                        is_silent=is_silent,
                        is_timed=is_timed,
                        total_steps=total_steps,
                        is_children_silent=is_children_silent,
                        is_enumerate=False)

    def enum(self, name, iterable: typing.Sized, *,
             is_silent: bool,
             is_children_silent: bool,
             is_timed: bool):
        return Iterator(logger=self,
                        name=name,
                        iterable=iterable,
                        is_silent=is_silent,
                        is_timed=is_timed,
                        total_steps=None,
                        is_children_silent=is_children_silent,
                        is_enumerate=True)

    def section(self, name, *,
                is_silent: bool,
                is_timed: bool,
                is_partial: bool,
                is_new_line: bool,
                is_children_silent: bool,
                total_steps: float):

        if self.__is_looping():
            if len(self.__sections) != 0:
                is_silent = True

            section = self.__loop.get_section(name=name,
                                              is_silent=is_silent,
                                              is_timed=is_timed,
                                              is_partial=is_partial,
                                              total_steps=total_steps,
                                              parents=[s.name for s in self.__sections])
            self.__sections.append(section)
        else:
            if len(self.__sections) > 0:
                if self.__sections[-1].is_silent or self.__sections[-1].is_children_silent:
                    is_silent = True
                    is_children_silent = True
            self.__sections.append(OuterSection(logger=self,
                                                name=name,
                                                is_silent=is_silent,
                                                is_timed=is_timed,
                                                is_partial=is_partial,
                                                is_new_line=is_new_line,
                                                is_children_silent=is_children_silent,
                                                total_steps=total_steps,
                                                level=len(self.__sections)))

        return self.__sections[-1]

    def progress(self, steps: float):
        if len(self.__sections) == 0:
            raise RuntimeError("You must be within a section to report progress")

        if self.__sections[-1].progress(steps):
            self.__log_line()

    def set_successful(self, is_successful=True):
        if len(self.__sections) == 0:
            raise RuntimeError("You must be within a section to report success")

        self.__sections[-1].is_successful = is_successful
        self.__log_line()

    def loop(self, iterator_: typing.Collection, *,
             is_track: bool,
             is_print_iteration_time: bool):
        if len(self.__sections) != 0:
            raise RuntimeError("Cannot start a loop within a section")

        self.__loop = Loop(iterator=iterator_,
                           logger=self,
                           is_track=is_track,
                           is_print_iteration_time=is_print_iteration_time)
        return self.__loop

    def finish_loop(self):
        if len(self.__sections) != 0:
            raise RuntimeError("Cannot be within a section when finishing the loop")
        self.__last_global_step = self.global_step
        self.__loop = None
        for w in self.__writers:
            w.flush()

    def section_enter(self, section):
        if len(self.__sections) == 0:
            raise RuntimeError("Entering a section without creating a section.\n"
                               "Always use logger.section to create a section")

        if section is not self.__sections[-1]:
            raise RuntimeError("Entering a section other than the one last_created\n"
                               "Always user with logger.section(...):")

        if len(self.__sections) > 1 and not self.__sections[-2].is_parented:
            self.__sections[-2].make_parent()
            if not self.__sections[-1].is_silent:
                self.log([])

        self.__log_line()

    def __log_looping_line(self):
        parts = [(f"{self.global_step :8,}:  ", Text.highlight)]
        parts += self.__loop.log_sections()
        parts += self.__indicators_print
        parts += self.__loop.log_progress()

        self.log(parts, is_new_line=False)

    def __log_line(self):
        if self.__is_looping():
            self.__log_looping_line()
            return

        if len(self.__sections) == 0:
            return

        parts = self.__sections[-1].log()
        if parts is None:
            return

        self.log(parts, is_new_line=False)

    def section_exit(self, section):
        if len(self.__sections) == 0:
            raise RuntimeError("Impossible")

        if section is not self.__sections[-1]:
            raise RuntimeError("Impossible")

        self.__log_line()
        self.__sections.pop(-1)

    def info(self, *args, **kwargs):
        self.__inspect.info(*args, **kwargs)


_internal: Optional[Logger] = None


def logger_singleton() -> Logger:
    global _internal
    if _internal is None:
        _internal = Logger()

    return _internal



import ast
import inspect
import textwrap
from typing import Callable, cast, Dict, Set


class DependencyParser(ast.NodeVisitor):
    secondary_attributes: Dict[str, Set[str]]

    def __init__(self, func: Callable):
        if type(func) == type:
            func = cast(object, func).__init__
            spec: inspect.Signature = inspect.signature(func)
            params = spec.parameters
            assert len(params) == 2
            param: inspect.Parameter = params[list(params.keys())[1]]
            source = textwrap.dedent(inspect.getsource(func))

        else:
            spec: inspect.Signature = inspect.signature(func)
            params = spec.parameters
            assert len(params) >= 1
            param: inspect.Parameter = params[list(params.keys())[0]]
            source = inspect.getsource(func)

        assert (param.kind == param.POSITIONAL_ONLY or
                param.kind == param.POSITIONAL_OR_KEYWORD)

        self.arg_name = param.name

        self.required = set()
        self.secondary_attributes = {}
        self.is_referenced = False

        source = textwrap.dedent(source)
        parsed = ast.parse(source)
        self.visit(parsed)

    def visit_Attribute(self, node: ast.Attribute):
        if isinstance(node.value, ast.Name) and node.value.id == self.arg_name:
            self.required.add(node.attr)
        elif (isinstance(node.value, ast.Attribute) and
              isinstance(node.value.value, ast.Name) and
              node.value.value.id == self.arg_name):
            self.required.add(node.value.attr)
            if node.value.attr not in self.secondary_attributes:
                self.secondary_attributes[node.value.attr] = set()
            self.secondary_attributes[node.value.attr].add(node.attr)
        else:
            for child in ast.iter_child_nodes(node):
                self.visit(child)

    def visit_Name(self, node: ast.Name):
        if node.id == self.arg_name:
            self.is_referenced = True
            print(f"Referenced {node.id} in {node.lineno}:{node.col_offset}")



import time

from labml import experiment
from labml_helpers.training_loop import TrainingLoop


def main():
    experiment.create(name='watcher')
    loop = TrainingLoop(loop_count=10, loop_step=1, is_save_models=False, log_new_line_interval=1,
                        log_write_interval=1, save_models_interval=1, is_loop_on_interrupt=False)
    with experiment.start():
        for i in loop:
            # time.sleep(10)
            print(i)
            # if i == 5:
            #     raise RuntimeError('oops')


if __name__ == '__main__':
    main()



import warnings
from collections import OrderedDict
from typing import List, Dict, Type, Set, Union
from typing import TYPE_CHECKING

from labml.utils.errors import ConfigsError

from labml import logger
from labml.logger import Text
from .config_function import ConfigFunction
from .config_item import ConfigItem
from .eval_function import EvalFunction
from .setup_function import SetupFunction

if TYPE_CHECKING:
    from .base import Configs

RESERVED = {'calc', 'list', 'set_hyperparams', 'set_meta', 'aggregate', 'calc_wrap', 'setup'}
_STANDARD_TYPES = {int, str, bool, float, Dict, List}


class PropertyKeys:
    calculators = '_calculators'
    evaluators = '_evaluators'
    hyperparams = '_hyperparams'
    aggregates = '_aggregates'
    meta = '_meta'
    setups = '_setups'


def _get_base_classes(class_: Type['Configs']) -> List[Type['Configs']]:
    classes = [class_]
    level = [class_]
    next_level = []

    while len(level) > 0:
        for c in level:
            for b in c.__bases__:
                if b == object:
                    continue
                next_level.append(b)
        classes += next_level
        level = next_level
        next_level = []

    classes.reverse()

    unique_classes = []
    hashes: Set[int] = set()
    for c in classes:
        if hash(c) not in hashes:
            unique_classes.append(c)
        hashes.add(hash(c))

    return unique_classes


class Parser:
    config_items: Dict[str, ConfigItem]
    options: Dict[str, Dict[str, Union[ConfigFunction, SetupFunction]]]
    evals: Dict[str, Dict[str, EvalFunction]]
    types: Dict[str, Type]
    values: Dict[str, any]
    explicitly_specified: Set[str]
    hyperparams: Dict[str, bool]
    meta: Dict[str, bool]
    aggregates: Dict[str, Dict[str, Dict[str, str]]]
    aggregate_parent: Dict[str, str]
    secondary_values: Dict[str, Dict[str, any]]

    def __init__(self, configs: 'Configs', values: Dict[str, any] = None, *,
                 is_directly_specified: bool):
        classes = _get_base_classes(type(configs))

        self.values = {}
        self.types = {}
        self.options = {}
        self.evals = {}
        self.config_items = {}
        self.configs = configs
        self.explicitly_specified = set()
        self.hyperparams = {}
        self.meta = {}
        self.aggregates = {}
        self.aggregate_parent = {}
        self.secondary_values = {}

        for c in classes:
            # for k, v in c.__annotations__.items():
            #     self.__collect_annotation(k, v)
            #
            for k, v in c.__dict__.items():
                if (PropertyKeys.evaluators in c.__dict__ and
                        k in c.__dict__[PropertyKeys.evaluators]):
                    continue
                self.__collect_config_item(k, v)

        for c in classes:
            if PropertyKeys.calculators in c.__dict__:
                for k, calculators in c.__dict__[PropertyKeys.calculators].items():
                    assert k in self.types, \
                        f"{k} calculator is present but the config declaration is missing"
                    for v in calculators:
                        self.__collect_calculator(k, v)

        for c in classes:
            if PropertyKeys.setups in c.__dict__:
                for k, setups in c.__dict__[PropertyKeys.setups].items():
                    if k not in self.types:
                        raise RuntimeError(f"{k} calculator is present but the config declaration is missing")
                    for v in setups:
                        self.__collect_setup(k, v)

        for c in classes:
            if PropertyKeys.evaluators in c.__dict__:
                for k, evaluators in c.__dict__[PropertyKeys.evaluators].items():
                    for v in evaluators:
                        self.__collect_evaluator(k, v)

        for c in classes:
            if PropertyKeys.hyperparams in c.__dict__:
                for k, is_hyperparam in c.__dict__[PropertyKeys.hyperparams].items():
                    self.hyperparams[k] = is_hyperparam

        for c in classes:
            if PropertyKeys.meta in c.__dict__:
                for k, is_meta in c.__dict__[PropertyKeys.meta].items():
                    self.meta[k] = is_meta

        for c in classes:
            if PropertyKeys.aggregates in c.__dict__:
                for k, aggregates in c.__dict__[PropertyKeys.aggregates].items():
                    self.aggregates[k] = aggregates

        for k, v in configs.__dict__.items():
            if k.startswith('_'):
                continue

            if k not in self.types:
                raise RuntimeError(f"Unknown key :{k}")
            self.__collect_value(k, v)

        if not is_directly_specified:
            self.explicitly_specified = set()

        if values is not None:
            for k, v in values.items():
                if k in self.types:
                    self.__collect_value(k, v)
                else:
                    parts = k.split('.')
                    if parts[0] in self.types:
                        if parts[0] not in self.secondary_values:
                            self.secondary_values[parts[0]] = {}
                        self.secondary_values[parts[0]][k[len(parts[0]) + 1:]] = v
                    else:
                        logger.log(f'Ignoring config: {k} = {str(v)}', Text.warning)

        self.__calculate_aggregates()
        self.__calculate_missing_values()

    @staticmethod
    def is_valid(key):
        if key.startswith('_'):
            return False

        if key in RESERVED:
            return False

        return True

    def __collect_config_item(self, k, v: ConfigItem):
        if not self.is_valid(k):
            return

        if v.has_value:
            self.values[k] = v.value

        if k in self.config_items:
            self.config_items[k].update(v)
        else:
            self.config_items[k] = v

        if k not in self.types:
            self.types[k] = v.annotation

    def __collect_value(self, k, v):
        if not self.is_valid(k):
            return

        self.explicitly_specified.add(k)

        self.values[k] = v
        if k not in self.types:
            self.types[k] = type(v)

    def __collect_annotation(self, k, v):
        if not self.is_valid(k):
            return

        self.types[k] = v

    def __collect_calculator(self, k, v: ConfigFunction):
        if k not in self.options:
            self.options[k] = OrderedDict()
        if v.option_name in self.options[k]:
            if v != self.options[k][v.option_name]:
                warnings.warn(f"Overriding option for {k}: {v.option_name}", Warning, stacklevel=5)

        self.options[k][v.option_name] = v

    def __collect_setup(self, k, v: SetupFunction):
        if k not in self.options:
            self.options[k] = OrderedDict()
        if v.option_name in self.options[k]:
            if v != self.options[k][v.option_name]:
                warnings.warn(f"Overriding option for {k}: {v.option_name}", Warning, stacklevel=5)

        self.options[k][v.option_name] = v

    def __collect_evaluator(self, k, v: EvalFunction):
        if k not in self.evals:
            self.evals[k] = OrderedDict()

        self.evals[k]['default'] = v

    def __calculate_missing_values(self):
        for k in self.types:
            if k in self.values:
                continue

            if k in self.options:
                self.values[k] = next(iter(self.options[k].keys()))
                continue

            if type(self.types[k]) == type:
                if self.types[k] in _STANDARD_TYPES:
                    continue

                try:
                    config_function = ConfigFunction(self.types[k],
                                                     config_names=self.config_items[k],
                                                     option_name='from_type')
                except ConfigsError as e:
                    continue

                self.options[k] = OrderedDict()
                self.options[k]['from_type'] = config_function
                self.values[k] = 'from_type'
                continue

    def __calculate_aggregates(self):
        queue = []
        for k in self.aggregates:
            if k not in self.values or self.values[k] not in self.aggregates[k]:
                continue

            queue.append(k)

        while queue:
            k = queue.pop()
            assert k in self.values
            option = self.values[k]
            assert option in self.aggregates[k]
            pairs = self.aggregates[k][option]

            for name, opt in pairs.items():
                if name in self.values and self.values[name] != '__none__':
                    continue

                self.aggregate_parent[name] = k
                self.values[name] = opt

                if name in self.aggregates and opt in self.aggregates[name]:
                    queue.append(name)



from pathlib import PurePath
from typing import Dict, List, Optional

from labml.internal.logger.store.indicators import Indicator
from labml.internal.logger.store.indicators.artifacts import Artifact
from labml.internal.logger.store.indicators.numeric import Scalar

from .indicators.factory import load_indicator_from_dict, create_default_indicator
from .namespace import Namespace
from ..writers import Writer
from ... import util
from ...lab import lab_singleton, LabYamlNotfoundError
from ...util import strings


class Store:
    dot_indicators: Dict[str, Indicator]
    namespaces: List[Namespace]
    indicators: Dict[str, Indicator]

    def __init__(self):
        self.indicators = {}
        self.dot_indicators = {}
        self.__indicators_file = None
        self.namespaces = []
        try:
            for ind in lab_singleton().indicators:
                self.add_indicator(load_indicator_from_dict(ind))
        except LabYamlNotfoundError:
            pass

        self.is_indicators_updated = True

    def save_indicators(self, writers: List[Writer], file: Optional[PurePath] = None):
        if not self.is_indicators_updated:
            return
        self.is_indicators_updated = False
        if file is None:
            if self.__indicators_file is None:
                return
            file = self.__indicators_file
        else:
            self.__indicators_file = file

        wildcards = {k: ind.to_dict() for k, ind in self.dot_indicators.items()}
        inds = {k: ind.to_dict() for k, ind in self.indicators.items()}
        with open(str(file), "w") as file:
            file.write(util.yaml_dump({'wildcards': wildcards,
                                       'indicators': inds}))

        for w in writers:
            w.save_indicators(self.dot_indicators, self.indicators)

    def __assert_name(self, name: str, value: any):
        if name.endswith("."):
            if name in self.dot_indicators:
                assert self.dot_indicators[name].equals(value)

        assert name not in self.indicators, f"{name} already used"

    def namespace_enter(self, ns: Namespace):
        self.namespaces.append(ns)

    def namespace_exit(self, ns: Namespace):
        if len(self.namespaces) == 0:
            raise RuntimeError("Impossible")

        if ns is not self.namespaces[-1]:
            raise RuntimeError("Impossible")

        self.namespaces.pop(-1)

    def add_indicator(self, indicator: Indicator):
        self.dot_indicators[indicator.name] = indicator
        self.is_indicators_updated = True

    def _create_indicator(self, key: str, value: any):
        if key in self.indicators:
            return

        ind_key, ind_score = strings.find_best_pattern(key, self.dot_indicators.keys())
        if ind_key is None:
            raise ValueError(f"Cannot find matching indicator for {key}")
        if ind_score == 0:
            is_print = self.dot_indicators[ind_key].is_print
            self.indicators[key] = create_default_indicator(key, value, is_print)
        else:
            self.indicators[key] = self.dot_indicators[ind_key].copy(key)
        self.is_indicators_updated = True

    def store(self, key: str, value: any):
        if key.endswith('.'):
            key = '.'.join([key[:-1]] + [ns.name for ns in self.namespaces])

        self._create_indicator(key, value)
        self.indicators[key].collect_value(value)

    def clear(self):
        for k, v in self.indicators.items():
            v.clear()

    def write(self, writer: Writer, global_step):
        return writer.write(global_step=global_step,
                            indicators=self.indicators)

    def create_namespace(self, name: str):
        return Namespace(store=self,
                         name=name)



import torch

from labml.logger import inspect


def test():
    inspect("test")
    inspect([1, 2])
    inspect({'x': 2, 4: '5'})
    small_torch = torch.arange(0, 10)
    large_torch = torch.arange(0, 1000).view(10, 100)
    small_np = small_torch.numpy()
    large_np = large_torch.numpy()
    inspect(small_np)
    inspect(large_np)
    inspect({'small': small_np, 'large': large_np})
    inspect({'small': small_torch, 'large': large_torch})
    inspect(small_torch)

    inspect(torch.tensor(small_np))


if __name__ == '__main__':
    test()



import inspect
import warnings
from typing import List, Callable, Set, Union, Optional

from labml.internal.configs.config_item import ConfigItem

from labml.internal.configs.dependency_parser import DependencyParser


class SetupFunction:
    func: Callable
    dependencies: Set[str]
    config_name: str

    def __check_type(self):
        key, pos = 0, 0
        spec: inspect.Signature = inspect.signature(self.func)
        params: List[inspect.Parameter] = list(spec.parameters.values())

        for p in params:
            if p.kind == p.POSITIONAL_OR_KEYWORD:
                pos += 1
            elif p.kind == p.KEYWORD_ONLY:
                key += 1
            else:
                raise RuntimeError(f"Only positional or keyword only arguments should be accepted: "
                                   f"{self.config_name} - {self.func.__name__}")

        assert pos >= 1

    def __get_dependencies(self):
        parser = DependencyParser(self.func)
        if parser.is_referenced:
            raise RuntimeError(f"{self.func.__name__} should only use attributes of configs")
        required = set(parser.required)

        for c in self.config_names:
            if c in required:
                required.remove(c)

        return required

    def __get_config_names(self, config_names: Union[str, ConfigItem, List[ConfigItem], List[str]]):
        if config_names is None:
            warnings.warn(f"Decorate your function with @setup(Config.{self.func.__name__})", FutureWarning, 4)
            return self.func.__name__
        elif type(config_names) == str:
            if self.check_string_names:
                warnings.warn(f"Use Config.{config_names} instead of '{config_names}'.", FutureWarning, 4)
            return config_names
        elif type(config_names) == ConfigItem:
            return config_names.key
        else:
            assert type(config_names) == list
            assert len(config_names) > 0
            keys = []
            for c in config_names:
                if isinstance(c, str):
                    warnings.warn(f"Use Config.{c} instead of '{c}'", FutureWarning, 4)
                    keys.append(c)
                elif isinstance(c, ConfigItem):
                    keys.append(c.key)
                else:
                    raise ValueError('Cannot determine config item', c)
            return keys

    def __get_option_name(self, option_name: str):
        if option_name is not None:
            return option_name
        else:
            return self.func.__name__

    def __init__(self, func, *,
                 option_name: Optional[str] = None,
                 config_names: Union[str, ConfigItem, List[ConfigItem], List[str]]):
        self.secondary_attributes = {}
        self.func = func
        self.option_name = self.__get_option_name(option_name)
        self.check_string_names = True
        self.config_names = self.__get_config_names(config_names)

        self.__check_type()
        self.dependencies = self.__get_dependencies()

    def __call__(self, configs: 'Configs'):
        self.func(configs)



import json
import socket
import ssl
import threading
import time
import urllib.error
import urllib.request
import webbrowser
from queue import Queue
from typing import Dict, Optional

import numpy as np

from labml import logger
from labml.logger import Text
from labml.utils.notice import labml_notice
from . import Writer as WriteBase
from ..store.indicators import Indicator
from ..store.indicators.numeric import NumericIndicator
from ...lab import lab_singleton

MAX_BUFFER_SIZE = 1024
WARMUP_COMMITS = 5
TIMEOUT_SECONDS = 5


class WebApiThread(threading.Thread):
    def __init__(self, url: str, verify_connection: bool):
        super().__init__(daemon=False)
        self.verify_connection = verify_connection
        self.url = url
        self.queue = Queue()
        self.is_stopped = False
        self.warned_updating = False

    def push(self, data: any):
        self.queue.put(data)

    def stop(self):
        self.is_stopped = True
        self.push('done')

    def run(self):
        while True:
            data = self.queue.get()
            if data == 'done':
                if self.warned_updating:
                    logger.log('Finished updating LabML App.', Text.highlight)
                return
            else:
                if self.is_stopped and not self.warned_updating:
                    logger.log('Still updating LabML App, please wait for it to complete..', Text.highlight)
                    self.warned_updating = True
                self._process(data)

    def _process(self, data: Dict[str, any]):
        res = self._send(data['packet'])
        if 'accept' in data:
            data['accept'](res)

    def _send(self, data: Dict[str, any]):
        req = urllib.request.Request(self.url)
        req.add_header('Content-Type', 'application/json; charset=utf-8')
        data = json.dumps(data)
        data = data.encode('utf-8')
        req.add_header('Content-Length', str(len(data)))

        try:
            if self.verify_connection:
                response = urllib.request.urlopen(req, data, timeout=TIMEOUT_SECONDS)
            else:
                response = urllib.request.urlopen(req, data, timeout=TIMEOUT_SECONDS,
                                                  context=ssl._create_unverified_context())
            content = response.read().decode('utf-8')
            result = json.loads(content)

            if result['errors']:
                err_log = ['LabML App errors\n']
                for e in result['errors']:
                    err_log += [(e['error'] + ': ', Text.key),
                                (e['message'], Text.value)]
                labml_notice(err_log)
            return result.get('url', None)
        except urllib.error.HTTPError as e:
            labml_notice([f'Failed to send to {self.url}: ',
                          (str(e.code), Text.value),
                          '\n' + str(e.reason)])
            return None
        except urllib.error.URLError as e:
            labml_notice([f'Failed to connect to {self.url}\n',
                          str(e.reason)])
            return None
        except socket.timeout as e:
            labml_notice([f'{self.url} timeout\n',
                          str(e)])
            return None


class Writer(WriteBase):
    url: Optional[str]
    thread: Optional[WebApiThread]

    def __init__(self):
        super().__init__()

        self.scalars_cache = []
        self.last_committed = time.time()
        self.commits_count = 0
        self.indicators = {}
        self.run_uuid = None
        self.name = None
        self.comment = None
        self.configs = None
        self.web_api = lab_singleton().web_api
        self.state = None

        if self.web_api is not None:
            self.thread = WebApiThread(self.web_api.url, self.web_api.verify_connection)
        else:
            self.thread = None

    @staticmethod
    def _parse_key(key: str):
        return key

    def set_info(self, *,
                 run_uuid: str,
                 name: str,
                 comment: str):
        self.run_uuid = run_uuid
        self.name = name
        self.comment = comment

    def set_configs(self, configs: Dict[str, any]):
        self.configs = configs

    def start(self):
        if self.web_api is None:
            return

        data = {
            'run_uuid': self.run_uuid,
            'name': self.name,
            'comment': self.comment,
            'time': time.time(),
            'configs': {}
        }

        if self.configs is not None:
            for k, v in self.configs.items():
                data['configs'][k] = v

        self.last_committed = time.time()
        self.commits_count = 0
        self.thread.push({'packet': data,
                          'accept': self._started})
        if not self.thread.is_alive():
            self.thread.start()

    def save_indicators(self, dot_indicators: Dict[str, Indicator], indicators: Dict[str, Indicator]):
        if self.web_api is None:
            return

        wildcards = {k: ind.to_dict() for k, ind in dot_indicators.items()}
        inds = {k: ind.to_dict() for k, ind in indicators.items()}
        data = {
            'run_uuid': self.run_uuid,
            'wildcard_indicators': wildcards,
            'indicators': inds
        }

        self.thread.push({'packet': data})

    def _started(self, url):
        if url is None:
            return None

        logger.log([(url, Text.link_href), (' Monitor experiment', Text.link_title)])
        if self.web_api.open_browser:
            webbrowser.open(url)

    def _write_indicator(self, global_step: int, indicator: Indicator):
        if indicator.is_empty():
            return

        if isinstance(indicator, NumericIndicator):
            value = indicator.get_mean()
            key = self._parse_key(indicator.mean_key)
            if key not in self.indicators:
                self.indicators[key] = []

            self.indicators[key].append((global_step, value))

    def write(self, *,
              global_step: int,
              indicators: Dict[str, Indicator]):
        if self.web_api is None:
            return

        for ind in indicators.values():
            self._write_indicator(global_step, ind)

        t = time.time()
        freq = self.web_api.frequency
        if self.commits_count < WARMUP_COMMITS:
            freq /= 2 ** (WARMUP_COMMITS - self.commits_count)
        if t - self.last_committed > freq:
            self.last_committed = t
            self.commits_count += 1
            self.flush()

    def status(self, status: str, details: str, time_: float):
        if self.web_api is None:
            return

        self.state = {
            'status': status,
            'details': details,
            'time': time_
        }

        self.flush()

        # TODO: Will have to fix this when there are other statuses that dont stop the experiment
        self.thread.stop()

    def flush(self):
        if self.web_api is None:
            return

        data = {}

        for key, value in self.indicators.items():
            value = np.array(value)
            step: np.ndarray = value[:, 0]
            value: np.ndarray = value[:, 1]
            while value.shape[0] > MAX_BUFFER_SIZE:
                if value.shape[0] % 2 == 1:
                    value = np.concatenate((value, value[-1:]))
                    step = np.concatenate((step, step[-1:]))

                n = value.shape[0] // 2
                step = np.mean(step.reshape(n, 2), axis=-1)
                value = np.mean(value.reshape(n, 2), axis=-1)

            data[key] = {
                'step': step.tolist(),
                'value': value.tolist()
            }

        self.indicators = {}

        self.thread.push({'packet': {
            'run_uuid': self.run_uuid,
            'track': data,
            'status': self.state,
            'time': time.time()
        }})



import inspect
from typing import List, Callable, Set

from labml.internal.configs.dependency_parser import DependencyParser


class EvalFunction:
    func: Callable
    dependencies: Set[str]
    config_name: str

    def __check_type(self):
        key, pos = 0, 0
        spec: inspect.Signature = inspect.signature(self.func)
        params: List[inspect.Parameter] = list(spec.parameters.values())

        for p in params:
            if p.kind == p.POSITIONAL_OR_KEYWORD:
                pos += 1
            elif p.kind == p.KEYWORD_ONLY:
                key += 1
            else:
                raise RuntimeError(f"Only positional or keyword only arguments should be accepted: "
                                   f"{self.config_name} - {self.func.__name__}")

        assert pos >= 1

    def __get_dependencies(self):
        parser = DependencyParser(self.func)
        if parser.is_referenced:
            raise RuntimeError(f"{self.func.__name__} should only use attributes of configs")
        return parser.required

    def __init__(self, func, *,
                 config_name: str):
        self.func = func
        self.config_name = config_name

        self.__check_type()
        self.dependencies = self.__get_dependencies()



import numpy as np
import torch

from labml.configs import BaseConfigs, option


class SeedConfigs(BaseConfigs):
    seed: int = 5

    set_seed = 'set_seed'

    def __init__(self):
        super().__init__(_primary='set_seed')


@option(SeedConfigs.set_seed)
def set_seed(c: SeedConfigs):
    torch.manual_seed(c.seed)
    np.random.seed(c.seed)
    return c.seed



import typing

from labml.utils.pytorch import store_model_indicators

from . import Indicator

if typing.TYPE_CHECKING:
    import torch


class PyTorchModule(Indicator):
    def __init__(self, name: str, is_print=False):
        super().__init__(name=name, is_print=is_print)

    def is_empty(self) -> bool:
        return True

    def collect_value(self, model: 'torch.nn.Module'):
        store_model_indicators(model, self.name)

    def get_histogram(self):
        return None

    def copy(self, key: str):
        return PyTorchModule(key, is_print=self.is_print)



from typing import Optional, Callable, Dict, Tuple

import numpy as np
import torch.optim
import torch.utils.data
from torch import nn

import labml.utils.pytorch as pytorch_utils
from labml import tracker, monit
from labml.configs import option, meta_config
from .training_loop import TrainingLoopConfigs
from labml.utils.pytorch import get_device


class BatchStepProtocol:
    def prepare_for_iteration(self) -> None:
        raise NotImplementedError()

    def init_stats(self) -> Dict[str, any]:
        return {}

    def update_stats(self, stats: Dict[str, any], update: Dict[str, any]):
        for k, v in update.items():
            if k not in stats:
                stats[k] = []
            stats[k].append(v)

    def log_stats(self, stats: Dict[str, any]):
        pass

    def process(self, batch: any, state: any) -> Tuple[Dict[str, any], any]:
        """
        Returns a tuple of stats and state
        """
        raise NotImplementedError()

    def update(self):
        pass


class BatchStep(BatchStepProtocol):
    def __init__(self, *,
                 model: nn.Module,
                 optimizer: torch.optim.Adam,
                 loss_func: Callable,
                 accuracy_func: Callable):
        self.accuracy_func = accuracy_func
        self.loss_func = loss_func
        self.optimizer = optimizer
        self.model = model
        hook_model_outputs(self.model)

        tracker.set_queue("loss.*", 20, True)
        tracker.set_scalar("accuracy.*", True)

    def log_stats(self, stats: any):
        if 'correct' in stats:
            tracker.add("accuracy.", np.sum(stats['correct']) / np.sum(stats['samples']))

    def prepare_for_iteration(self):
        if MODE_STATE.is_train:
            self.model.train()
        else:
            self.model.eval()

    def calculate_loss(self, batch: any, state: any):
        device = get_device(self.model)
        data, target = batch
        data, target = data.to(device), target.to(device)
        stats = {
            'samples': len(data)
        }

        with monit.section("model"):
            output = self.model(data)

        if isinstance(output, tuple):
            output = output[0]

        loss = self.loss_func(output, target)
        if self.accuracy_func is not None:
            stats['correct'] = self.accuracy_func(output, target)

        tracker.add("loss.", loss)

        return loss, stats, None

    def process(self, batch: any, state: any):
        loss, stats, new_state = self.calculate_loss(batch, state)

        if MODE_STATE.is_train:
            with monit.section('backward'):
                loss.backward()

        return stats, new_state

    def update(self):
        if not MODE_STATE.is_train:
            return
        with monit.section('optimize'):
            self.optimizer.step()
        if MODE_STATE.is_log_parameters:
            pytorch_utils.store_model_indicators(self.model)
        self.optimizer.zero_grad()


class ModeState:
    def __init__(self):
        self._rollback_stack = []

        self.is_log_activations = False
        self.is_train = False
        self.is_log_parameters = False

    def enter(self, mode: Dict[str, any]):
        rollback = {}
        for k, v in mode.items():
            if v is None:
                continue
            rollback[k] = getattr(self, k)
            setattr(self, k, v)

        self._rollback_stack.append(rollback)

        return len(self._rollback_stack)

    def exit(self, n: int):
        assert n == len(self._rollback_stack)

        rollback = self._rollback_stack[-1]
        self._rollback_stack.pop(-1)

        for k, v in rollback.items():
            setattr(self, k, v)


MODE_STATE = ModeState()


class Mode:
    def __init__(self, *,
                 is_train: bool = None,
                 is_log_parameters: bool = None,
                 is_log_activations: bool = None):
        self.mode = {}
        if is_log_activations is not None:
            self.mode['is_log_activations'] = is_log_activations
        if is_train is not None:
            self.mode['is_train'] = is_train
        if is_log_parameters is not None:
            self.mode['is_log_parameters'] = is_log_parameters
        # for k, v in kwargs.items():
        #     if k[0] == '_' or not hasattr(_MODE_STATE, k):
        #         raise RuntimeError(f"Unknown mode {k}={v}")
        #
        #     self.mode[k] = v

        self.idx = -1

    def __enter__(self):
        self.idx = MODE_STATE.enter(self.mode)

    def __exit__(self, exc_type, exc_val, exc_tb):
        MODE_STATE.exit(self.idx)


class ForwardHook:
    def __init__(self, model_name, name: str, module: torch.nn.Module):
        self.model_name = model_name
        self.name = name
        self.module = module
        module.register_forward_hook(self)

    def save(self, name: str, output):
        if isinstance(output, torch.Tensor):
            pytorch_utils.store_l1_l2(name, output)
        elif isinstance(output, tuple):
            for i, o in enumerate(output):
                self.save(f"{name}.{i}", o)

    def __call__(self, module, i, o):
        if not MODE_STATE.is_log_activations:
            return

        self.save(f"module.{self.model_name}.{self.name}", o)


def hook_model_outputs(model: torch.nn.Module, model_name: str = "model"):
    for name, module in model.named_modules():
        if name == '':
            name = 'full'
        ForwardHook(model_name, name, module)


class Trainer:
    def __init__(self, *,
                 name: str,
                 batch_step: BatchStepProtocol,
                 data_loader: torch.utils.data.DataLoader,
                 is_increment_global_step: bool,
                 log_interval: Optional[int],
                 update_interval: Optional[int],
                 inner_iterations: int):
        self.batch_step = batch_step
        self.log_interval = log_interval
        self.update_interval = update_interval
        self.is_increment_global_step = is_increment_global_step
        self.data_loader = data_loader
        self.name = name
        self.__total_steps = len(data_loader)
        self.__iteration_idx = -1
        self.__iterable = None
        self.__n_iteration = -1
        self.inner_iterations = inner_iterations
        self.__state = None
        self.__stats = {}

    def __call__(self):
        if self.__iterable is None or self.__n_iteration >= self.inner_iterations:
            self.__iterable = iter(self.data_loader)
            self.__iteration_idx = 0
            self.__n_iteration = 0
            self.stats = self.batch_step.init_stats()
        self.batch_step.prepare_for_iteration()
        self.__state = None
        with torch.set_grad_enabled(MODE_STATE.is_train):
            self.__iterate()

    def __iterate(self):
        self.__n_iteration += 1
        if self.__iteration_idx >= self.__n_iteration * self.__total_steps / self.inner_iterations:
            return

        is_updated = True

        partial_count = 0
        with monit.section(self.name, is_partial=True):
            if self.__iteration_idx == 0:
                monit.progress(0)
            while True:
                i = self.__iteration_idx
                batch = next(self.__iterable)

                with Mode(is_log_activations=(MODE_STATE.is_log_activations and partial_count == 0)):
                    update, self.__state = self.batch_step.process(batch, self.__state)

                is_updated = False
                self.batch_step.update_stats(self.stats, update)

                if self.is_increment_global_step:
                    tracker.add_global_step(update['samples'])

                if self.update_interval is not None and (i + 1) % self.update_interval == 0:
                    self.batch_step.update()
                    is_updated = True

                if self.log_interval is not None and (i + 1) % self.log_interval == 0:
                    tracker.save()

                self.__iteration_idx += 1
                partial_count += 1
                monit.progress(self.__iteration_idx / self.__total_steps)

                if self.__iteration_idx >= self.__n_iteration * self.__total_steps / self.inner_iterations:
                    break

        if not is_updated:
            self.batch_step.update()

        self.batch_step.log_stats(self.stats)


class TrainValidConfigs(TrainingLoopConfigs):
    epochs: int = 10

    loss_func: Callable
    accuracy_func: Callable
    optimizer: torch.optim.Adam
    model: nn.Module
    batch_step: BatchStepProtocol = 'simple_batch_step'

    trainer: Trainer
    validator: Trainer

    train_log_interval: int = 10
    train_update_interval: int = 1

    loop_count = 'data_loop_count'
    loop_step = None

    train_loader: torch.utils.data.DataLoader
    valid_loader: torch.utils.data.DataLoader

    inner_iterations: int = 1

    is_log_parameters: bool = True
    is_log_activations: bool = True

    def run_step(self):
        for i in range(self.inner_iterations):
            with tracker.namespace('sample'):
                self.sample()
            with Mode(is_train=True,
                      is_log_parameters=self.is_log_parameters,
                      is_log_activations=self.is_log_activations):
                with tracker.namespace('train'):
                    self.trainer()
            if self.validator:
                with tracker.namespace('valid'):
                    self.validator()

    def run(self):
        for _ in self.training_loop:
            self.run_step()

    def sample(self):
        pass


@option(TrainValidConfigs.batch_step)
def simple_batch_step(c: TrainValidConfigs):
    return BatchStep(model=c.model,
                     optimizer=c.optimizer,
                     loss_func=c.loss_func,
                     accuracy_func=c.accuracy_func)


@option(TrainValidConfigs.trainer)
def trainer(c: TrainValidConfigs):
    return Trainer(name='Train',
                   batch_step=c.batch_step,
                   data_loader=c.train_loader,
                   is_increment_global_step=True,
                   log_interval=c.train_log_interval,
                   update_interval=c.train_update_interval,
                   inner_iterations=c.inner_iterations)


@option(TrainValidConfigs.validator)
def validator(c: TrainValidConfigs):
    return Trainer(name='Valid',
                   batch_step=c.batch_step,
                   data_loader=c.valid_loader,
                   is_increment_global_step=False,
                   log_interval=None,
                   update_interval=None,
                   inner_iterations=c.inner_iterations)


@option(TrainValidConfigs.optimizer)
def optimizer(c: TrainValidConfigs):
    from labml_helpers.optimizer import OptimizerConfigs
    opt_conf = OptimizerConfigs()
    opt_conf.parameters = c.model.parameters()
    return opt_conf


@option(TrainValidConfigs.loop_count)
def data_loop_count(c: TrainValidConfigs):
    return c.epochs


meta_config(TrainValidConfigs.train_log_interval,
            TrainValidConfigs.train_update_interval,
            TrainValidConfigs.is_log_parameters,
            TrainValidConfigs.is_log_activations)



import torch

from labml import tracker
from labml.configs import BaseConfigs


def store_l1_l2(name: str, tensor: torch.Tensor):
    if tensor.is_floating_point():
        tracker.add(f"{name}.mean", tensor.mean())
        tracker.add(f"{name}.l1", tensor.abs().mean())
        tracker.add(f"{name}.l2", (tensor ** 2).mean().sqrt())


def store_model_indicators(model: torch.nn.Module, model_name: str = "model"):
    for name, param in model.named_parameters():
        if param.requires_grad:
            with torch.no_grad():
                store_l1_l2(f"param.{model_name}.{name}", param)
                if param.grad is not None:
                    store_l1_l2(f"grad.{model_name}.{name}", param.grad)


def get_modules(configs: BaseConfigs):
    keys = dir(configs)

    modules = {}
    for k in keys:
        value = getattr(configs, k)
        if isinstance(value, torch.nn.Module):
            modules[k] = value

    return modules


def get_device(module: torch.nn.Module):
    params = module.parameters()
    try:
        sample_param = next(params)
        return sample_param.device
    except StopIteration:
        raise RuntimeError(f"Unable to determine"
                           f" device of {module.__class__.__name__}") from None



import json
import os
import pathlib
import time
from typing import Optional, List, Set, Dict, Union

import git

from labml import logger, monit
from labml.internal.configs.base import Configs
from labml.internal.configs.processor import ConfigProcessor
from labml.internal.configs.processor_dict import ConfigProcessorDict
from labml.internal.experiment.experiment_run import Run, struct_time_to_time, struct_time_to_date
from labml.internal.experiment.watcher import ExperimentWatcher
from labml.internal.lab import lab_singleton
from labml.internal.logger import logger_singleton as logger_internal
from labml.internal.util import is_ipynb, is_colab, is_kaggle
from labml.logger import Text
from labml.utils import get_caller_file
from labml.utils.notice import labml_notice


class ModelSaver:
    def save(self, checkpoint_path: pathlib.Path) -> any:
        raise NotImplementedError()

    def load(self, checkpoint_path: pathlib.Path, info: any):
        raise NotImplementedError()


class CheckpointSaver:
    model_savers: Dict[str, ModelSaver]

    def __init__(self, path: pathlib.PurePath):
        self.path = path
        self.model_savers = {}
        self.__no_savers_warned = False

    def add_savers(self, models: Dict[str, ModelSaver]):
        """
        ## Set variable for saving and loading
        """
        self.model_savers.update(models)

    def save(self, global_step):
        """
        ## Save model as a set of numpy arrays
        """

        if not self.model_savers:
            if not self.__no_savers_warned:
                labml_notice(["No models were registered for saving\n",
                              "You can register models with ",
                              ('experiment.add_pytorch_models', Text.value)])
                self.__no_savers_warned = True
            return

        checkpoints_path = pathlib.Path(self.path)
        if not checkpoints_path.exists():
            checkpoints_path.mkdir()

        checkpoint_path = checkpoints_path / str(global_step)
        assert not checkpoint_path.exists()

        checkpoint_path.mkdir()

        info = {}
        for name, saver in self.model_savers.items():
            info[name] = saver.save(checkpoint_path)

        # Save header
        with open(str(checkpoint_path / "info.json"), "w") as f:
            f.write(json.dumps(info))

    def load(self, checkpoint_path: pathlib.Path, models: List[str] = None):
        """
        ## Load model as a set of numpy arrays
        """

        if not models:
            models = list(self.model_savers.keys())

        with open(str(checkpoint_path / "info.json"), "r") as f:
            info = json.loads(f.readline())

        to_load = []
        not_loaded = []
        missing = []
        for name in models:
            if name not in info:
                missing.append(name)
            else:
                to_load.append(name)
        for name in info:
            if name not in models:
                not_loaded.append(name)

        # Load each model
        for name in to_load:
            saver = self.model_savers[name]
            saver.load(checkpoint_path, info[name])

        if missing:
            labml_notice([(f'{missing} ', Text.highlight),
                          ('model(s) could not be found.\n'),
                          (f'{to_load} ', Text.none),
                          ('models were loaded.', Text.none)
                          ], is_danger=True)
        if not_loaded:
            labml_notice([(f'{not_loaded} ', Text.none),
                          ('models were not loaded.\n', Text.none),
                          'Models to be loaded should be specified with: ',
                          ('experiment.add_pytorch_models', Text.value)])


class Experiment:
    r"""
    Each experiment has different configurations or algorithms.
    An experiment can have multiple runs.

    Keyword Arguments:
        name (str, optional): name of the experiment
        python_file (str, optional): path of the Python file that
            created the experiment
        comment (str, optional): a short description of the experiment
        writers (Set[str], optional): list of writers to write stat to
        ignore_callers: (Set[str], optional): list of files to ignore when
            automatically determining ``python_file``
        tags (Set[str], optional): Set of tags for experiment
    """
    run: Run
    configs_processor: Optional[ConfigProcessor]

    # whether not to start the experiment if there are uncommitted changes.
    check_repo_dirty: bool
    checkpoint_saver: CheckpointSaver

    def __init__(self, *,
                 name: Optional[str],
                 python_file: Optional[str],
                 comment: Optional[str],
                 writers: Set[str],
                 ignore_callers: Set[str],
                 tags: Optional[Set[str]],
                 is_evaluate: bool):
        if python_file is None:
            python_file = get_caller_file(ignore_callers)

        if python_file.startswith('<ipython'):
            assert is_ipynb()
            if name is None:
                raise ValueError("You must specify python_file or experiment name"
                                 " when creating an experiment from a python notebook.")

            lab_singleton().set_path(os.getcwd())
            python_file = 'notebook.ipynb'
        else:
            lab_singleton().set_path(python_file)

            if name is None:
                file_path = pathlib.PurePath(python_file)
                name = file_path.stem

        if comment is None:
            comment = ''
        if global_params_singleton().comment is not None:
            comment = global_params_singleton().comment

        self.name = name
        self.experiment_path = lab_singleton().experiments / name

        self.check_repo_dirty = lab_singleton().check_repo_dirty

        self.configs_processor = None

        if tags is None:
            tags = set(name.split('_'))

        self.run = Run.create(
            experiment_path=self.experiment_path,
            python_file=python_file,
            trial_time=time.localtime(),
            comment=comment,
            tags=list(tags))

        try:
            repo = git.Repo(lab_singleton().path)

            self.run.commit = repo.head.commit.hexsha
            self.run.commit_message = repo.head.commit.message.strip()
            self.run.is_dirty = repo.is_dirty()
            self.run.diff = repo.git.diff()
        except git.InvalidGitRepositoryError:
            if not is_colab() and not is_kaggle():
                labml_notice(["Not a valid git repository: ",
                              (lab_singleton().path, Text.value)])
            self.run.commit = 'unknown'
            self.run.commit_message = ''
            self.run.is_dirty = True
            self.run.diff = ''

        logger_internal().reset_writers()

        if 'sqlite' in writers:
            from labml.internal.logger.writers import sqlite
            logger_internal().add_writer(
                sqlite.Writer(self.run.sqlite_path, self.run.artifacts_folder))
        if 'tensorboard' in writers:
            from labml.internal.logger.writers import tensorboard
            logger_internal().add_writer(tensorboard.Writer(self.run.tensorboard_log_path))
        if 'file' in writers:
            from labml.internal.logger.writers import file
            logger_internal().add_writer(file.Writer(self.run.log_file))
        if 'web_api' in writers:
            from labml.internal.logger.writers import web_api
            self.web_api = web_api.Writer()
            logger_internal().add_writer(self.web_api)
        else:
            self.web_api = None

        self.checkpoint_saver = CheckpointSaver(self.run.checkpoint_path)
        self.is_evaluate = is_evaluate

    def __print_info(self):
        """
        🖨 Print the experiment info and check git repo status
        """

        logger.log()
        logger.log([
            (self.name, Text.title),
            ': ',
            (str(self.run.uuid), Text.meta)
        ])

        if self.run.comment != '':
            logger.log(['\t', (self.run.comment, Text.highlight)])

        commit_message = self.run.commit_message.strip().replace('\n', '¶ ').replace('\r', '')
        logger.log([
            "\t"
            "[dirty]" if self.run.is_dirty else "[clean]",
            ": ",
            (f"\"{commit_message}\"", Text.highlight)
        ])

        if self.run.load_run is not None:
            logger.log([
                "\t"
                "loaded from",
                ": ",
                (f"{self.run.load_run}", Text.meta2),
            ])

    def _load_checkpoint(self, checkpoint_path: pathlib.Path):
        self.checkpoint_saver.load(checkpoint_path)

    def save_checkpoint(self):
        self.checkpoint_saver.save(logger_internal().global_step)

    def calc_configs(self,
                     configs: Configs,
                     configs_override: Optional[Dict[str, any]],
                     run_order: Optional[List[Union[List[str], str]]]):
        if configs_override is None:
            configs_override = {}
        if global_params_singleton().configs is not None:
            configs_override.update(global_params_singleton().configs)

        self.configs_processor = ConfigProcessor(configs, configs_override)
        self.configs_processor(run_order)

        logger.log()

    def calc_configs_dict(self,
                          configs: Dict[str, any],
                          configs_override: Optional[Dict[str, any]]):
        self.configs_processor = ConfigProcessorDict(configs, configs_override)
        self.configs_processor()

        logger.log()

    def __start_from_checkpoint(self, run_uuid: str, checkpoint: Optional[int]):
        checkpoint_path, global_step = experiment_run.get_run_checkpoint(
            run_uuid,
            checkpoint)

        if global_step is None:
            return 0
        else:
            with monit.section("Loading checkpoint"):
                self._load_checkpoint(checkpoint_path)
            self.run.load_run = run_uuid

        return global_step

    def load_models(self, *,
                    models: List[str],
                    run_uuid: Optional[str] = None,
                    checkpoint: Optional[int] = None):
        if checkpoint is None:
            checkpoint = -1
        checkpoint_path, global_step = experiment_run.get_run_checkpoint(
            run_uuid,
            checkpoint)

        if global_step is None:
            labml_notice(['Could not find saved checkpoint'], is_danger=True)
            return

        with monit.section("Loading checkpoint"):
            _ = self.checkpoint_saver.load(checkpoint_path, models)

    def start(self, *,
              run_uuid: Optional[str] = None,
              checkpoint: Optional[int] = None):
        if run_uuid is not None:
            if checkpoint is None:
                checkpoint = -1
            global_step = self.__start_from_checkpoint(run_uuid, checkpoint)
        else:
            global_step = 0

        self.run.start_step = global_step
        logger_internal().set_start_global_step(global_step)

        self.__print_info()
        if self.check_repo_dirty and self.run.is_dirty:
            logger.log([("[FAIL]", Text.danger),
                        " Cannot trial an experiment with uncommitted changes."])
            exit(1)

        if self.configs_processor is not None:
            self.configs_processor.print()

        if not self.is_evaluate:
            self.run.save_info()

            if self.configs_processor is not None:
                self.configs_processor.save(self.run.configs_path)

            logger_internal().save_indicators(self.run.indicators_path)

        if self.web_api is not None:
            self.web_api.set_info(run_uuid=self.run.uuid,
                                  name=self.name,
                                  comment=self.run.comment)
            if self.configs_processor is not None:
                self.web_api.set_configs(self.configs_processor.to_json())
            self.web_api.start()

        if self.configs_processor:
            logger_internal().write_h_parameters(self.configs_processor.get_hyperparams())

        return ExperimentWatcher(self)

    def finish(self, status: str, details: any = None):
        if not self.is_evaluate:
            with open(str(self.run.run_log_path), 'a') as f:
                end_time = time.time()
                data = json.dumps({'status': status,
                                   'details': details,
                                   'time': end_time}, indent=None)
                f.write(data + '\n')

        if self.web_api is not None:
            self.web_api.status(status, details, end_time)


class GlobalParams:
    def __init__(self):
        self.configs = None
        self.comment = None


_global_params: Optional[GlobalParams] = None
_internal: Optional[Experiment] = None


def global_params_singleton() -> GlobalParams:
    global _global_params

    if _global_params is None:
        _global_params = GlobalParams()

    return _global_params


def experiment_singleton() -> Experiment:
    global _internal

    assert _internal is not None

    return _internal


def create_experiment(*,
                      name: Optional[str],
                      python_file: Optional[str],
                      comment: Optional[str],
                      writers: Set[str],
                      ignore_callers: Set[str],
                      tags: Optional[Set[str]],
                      is_evaluate: bool):
    global _internal

    _internal = Experiment(name=name,
                           python_file=python_file,
                           comment=comment,
                           writers=writers,
                           ignore_callers=ignore_callers,
                           tags=tags,
                           is_evaluate=is_evaluate)



import types
from typing import Dict, List, Callable, Union, Tuple, Optional

from labml.internal.configs.eval_function import EvalFunction
from .config_function import ConfigFunction
from .config_item import ConfigItem
from .parser import Parser, PropertyKeys
from .setup_function import SetupFunction


def _is_class_method(func: Callable):
    if not isinstance(func, types.FunctionType):
        return False
    # if not callable(func):
    #     return False

    import inspect

    spec: inspect.Signature = inspect.signature(func)
    params: List[inspect.Parameter] = list(spec.parameters.values())
    if len(params) == 0:
        raise RuntimeError('Can only have methods in a config class', func)
    p = params[0]
    if p.kind != p.POSITIONAL_OR_KEYWORD:
        raise RuntimeError('Can only have methods in a config class', func)
    if p.name != 'self':
        raise RuntimeError('Can only have methods in a config class', func)

    return True


class Configs:
    _calculators: Dict[str, List[ConfigFunction]] = {}
    _evaluators: Dict[str, List[EvalFunction]] = {}
    _setups: Dict[str, List[SetupFunction]] = {}

    def __init__(self, *, _primary: str = None):
        self._primary = _primary

    def __init_subclass__(cls, **kwargs):
        configs = {}

        for k, v in cls.__annotations__.items():
            if not Parser.is_valid(k):
                continue

            configs[k] = ConfigItem(key=k,
                                    configs_class=cls,
                                    has_annotation=True, annotation=v,
                                    has_value=k in cls.__dict__,
                                    value=cls.__dict__.get(k, None))

        evals = []

        if PropertyKeys.setups not in cls.__dict__:
            cls._setups = {}

        for k, v in cls.__dict__.items():
            if not Parser.is_valid(k):
                continue

            if _is_class_method(v):
                evals.append((k, v))
                continue

            configs[k] = ConfigItem(key=k,
                                    configs_class=cls,
                                    has_annotation=k in cls.__annotations__,
                                    annotation=cls.__annotations__.get(k, None),
                                    has_value=True, value=v)

        for e in evals:
            cls._add_eval_function(e[1], e[0])

        for k, v in configs.items():
            setattr(cls, k, v)

    @classmethod
    def _add_config_function(cls,
                             func: Callable,
                             name: Union[ConfigItem, List[ConfigItem]],
                             option: Optional[str],
                             pass_params: Optional[List[ConfigItem]]):
        if PropertyKeys.calculators not in cls.__dict__:
            cls._calculators = {}

        calc = ConfigFunction(func,
                              config_names=name,
                              option_name=option,
                              pass_params=pass_params)
        if type(calc.config_names) == str:
            config_names = [calc.config_names]
        else:
            config_names = calc.config_names

        for n in config_names:
            if n not in cls._calculators:
                cls._calculators[n] = []
            cls._calculators[n].append(calc)

    @classmethod
    def _add_eval_function(cls,
                           func: Callable,
                           name: str):
        if PropertyKeys.evaluators not in cls.__dict__:
            cls._evaluators = {}

        calc = EvalFunction(func,
                            config_name=name)

        if name not in cls._evaluators:
            cls._evaluators[name] = []
        cls._evaluators[name].append(calc)

    @classmethod
    def _add_setup_function(cls,
                            func: Callable,
                            config_names: List[ConfigItem],
                            option_name: Optional[str]):
        if PropertyKeys.setups not in cls.__dict__:
            cls._setups = {}

        calc = SetupFunction(func,
                             option_name=option_name,
                             config_names=config_names)
        config_names = calc.config_names

        for n in config_names:
            if n not in cls._setups:
                cls._setups[n] = []
            cls._setups[n].append(calc)

    @classmethod
    def _calc(cls,
              name: Union[ConfigItem, List[ConfigItem]] = None,
              option: Optional[str] = None,
              pass_params: Optional[List[ConfigItem]] = None):

        def wrapper(func: Callable):
            cls._add_config_function(func, name, option, pass_params)

            return func

        return wrapper

    @classmethod
    def calc_wrap(cls, func: Callable,
                  name: Union[ConfigItem, List[ConfigItem]],
                  option_name: Optional[str] = None,
                  pass_params: Optional[List[ConfigItem]] = None):
        cls._add_config_function(func, name, option_name, pass_params)

        return func

    @classmethod
    def setup(cls, names: Union[List[ConfigItem], ConfigItem], option_name: Optional[str]):
        if not isinstance(names, list):
            names: List[ConfigItem] = [names]

        def wrapper(func: Callable):
            cls._add_setup_function(func, names, option_name)
            return func

        return wrapper

    @classmethod
    def calc(cls,
             name: Union[ConfigItem, List[ConfigItem]] = None,
             option_name: Optional[str] = None,
             pass_params: Optional[List[ConfigItem]] = None):
        return cls._calc(name, option_name, pass_params)

    @classmethod
    def set_hyperparams(cls, *args: ConfigItem, is_hyperparam=True):
        if PropertyKeys.hyperparams not in cls.__dict__:
            cls._hyperparams = {}

        for h in args:
            cls._hyperparams[h.key] = is_hyperparam

    @classmethod
    def set_meta(cls, *args: ConfigItem, is_meta=True):
        if PropertyKeys.meta not in cls.__dict__:
            cls._meta = {}

        for h in args:
            cls._meta[h.key] = is_meta

    @classmethod
    def aggregate(cls, name: ConfigItem, option: str,
                  *args: Tuple[ConfigItem, any]):
        assert args

        if PropertyKeys.aggregates not in cls.__dict__:
            cls._aggregates = {}

        if name.key not in cls._aggregates:
            cls._aggregates[name.key] = {}

        pairs = {p[0].key: p[1] for p in args}
        cls._aggregates[name.key][option] = pairs



from pathlib import PurePath, Path
from typing import Dict, Optional, List, Union

from labml import logger
from labml.internal import util
from labml.logger import Text
from .base import Configs
from .calculator import Calculator
from .parser import Parser
from .utils import Value

_CONFIG_PRINT_LEN = 40


class ConfigProcessor:
    def __init__(self, configs: Configs, values: Dict[str, any] = None, *,
                 is_directly_specified: bool = True):
        self.parser = Parser(configs, values, is_directly_specified=is_directly_specified)
        self.calculator = Calculator(configs=configs,
                                     options=self.parser.options,
                                     evals=self.parser.evals,
                                     types=self.parser.types,
                                     values=self.parser.values,
                                     secondary_values=self.parser.secondary_values,
                                     aggregate_parent=self.parser.aggregate_parent)

    def __call__(self, run_order: Optional[List[Union[List[str], str]]] = None):
        self.calculator(run_order)

    def save(self, configs_path: PurePath):
        configs = self.to_json()

        with open(str(configs_path), "w") as file:
            file.write(util.yaml_dump(configs))

    def _get_options_list(self, key: str):
        opts = list(self.parser.options.get(key, {}).keys())
        if not opts:
            opts = list((self.parser.aggregates.get(key, {}).keys()))

        return opts

    def to_json(self):
        orders = {k: i for i, k in enumerate(self.calculator.topological_order)}
        configs = {}
        for k, v in self.parser.types.items():
            configs[k] = {
                'name': k,
                'type': str(v),
                'value': Value.to_yaml_truncated(self.parser.values.get(k, None)),
                'order': orders.get(k, -1),
                'options': self._get_options_list(k),
                'computed': Value.to_yaml_truncated(getattr(self.calculator.configs, k, None)),
                'is_hyperparam': self.parser.hyperparams.get(k, None),
                'is_meta': self.parser.meta.get(k, None),
                'is_explicitly_specified': (k in self.parser.explicitly_specified)
            }

        for k, proc in self.calculator.config_processors.items():
            sub_configs = proc.to_json()
            for sk, v in sub_configs.items():
                configs[f"{k}.{sk}"] = v

        return configs

    def get_hyperparams(self):
        order = self.calculator.topological_order.copy()

        hyperparams = {}
        for key in order:
            if (self.parser.hyperparams.get(key, False) or
                    key in self.parser.explicitly_specified):
                value = getattr(self.calculator.configs, key, None)
                if key in self.parser.options:
                    value = self.parser.values[key]

                if type(value) not in {int, float, str}:
                    value = Value.to_str(value)

                hyperparams[key] = value

        return hyperparams

    def __print_config(self, key, *, indentation: int, value=None, option=None,
                       other_options=None, is_ignored=False, is_list=False):
        logger.log(self.__config_log_parts(key,
                                           indentation=indentation,
                                           value=value,
                                           option=option,
                                           other_options=other_options,
                                           is_ignored=is_ignored,
                                           is_list=is_list))
        if key in self.calculator.config_processors:
            self.calculator.config_processors[key].print_configs(indentation + 1)

    def __config_log_parts(self, key, *, indentation: int, value=None, option=None,
                           other_options=None, is_ignored=False, is_list=False):
        parts = ['\t' * indentation]

        if is_ignored:
            parts.append((key, Text.subtle))
            return parts

        is_hyperparam = self.parser.hyperparams.get(key, None)
        is_meta = self.parser.meta.get(key, None)
        if is_hyperparam is None:
            is_hyperparam = key in self.parser.explicitly_specified
            if is_meta:
                is_hyperparam = False

        if is_hyperparam:
            parts.append((key, [Text.key, Text.highlight]))
        elif is_meta:
            parts.append((key, Text.subtle))
        else:
            parts.append((key, Text.key))

        if is_list:
            parts.append(('[]', Text.subtle))

        parts.append((' = ', Text.subtle))

        if other_options is None:
            other_options = []

        if value is not None:
            value_str = Value.to_str(value)

            value_str = value_str.replace('\n', '')
            if len(value_str) < _CONFIG_PRINT_LEN:
                parts.append((f"{value_str}", Text.value))
            else:
                parts.append((f"{value_str[:_CONFIG_PRINT_LEN]}...", Text.value))
            parts.append('\t')

        if option is not None:
            if len(other_options) == 0:
                parts.append((option, Text.subtle))
            else:
                parts.append((option, Text.none))

        if value is None and option is None:
            parts.append(("None", Text.value))
            parts.append('\t')

        if len(other_options) > 0:
            parts.append(('\t[', Text.subtle))
            for i, opt in enumerate(other_options):
                if i > 0:
                    parts.append((', ', Text.subtle))
                parts.append(opt)
            parts.append((']', Text.subtle))

        return parts

    def print_configs(self, indentation: int):
        order = self.calculator.topological_order.copy()
        order.sort()
        added = set(order)
        ignored = set()

        for k in self.parser.types:
            if k not in added:
                added.add(k)
                order.append(k)
                ignored.add(k)

        for k in order:
            computed = getattr(self.calculator.configs, k, None)
            opts = self._get_options_list(k)
            if k in ignored:
                self.__print_config(k, indentation=indentation, is_ignored=True)
            elif opts:
                v = self.parser.values[k]
                if v in opts:
                    opts.remove(v)
                else:
                    v = None

                self.__print_config(k,
                                    indentation=indentation,
                                    value=computed,
                                    option=v,
                                    other_options=opts)
            else:
                self.__print_config(k, indentation=indentation, value=computed)

    def print(self):
        logger.log("Configs:", Text.heading)
        self.print_configs(1)
        logger.log()


def load_configs(configs_path: PurePath):
    configs_path = Path(configs_path)
    if not configs_path.exists():
        return None

    with open(str(configs_path), 'r') as file:
        configs = util.yaml_load(file.read())

    return configs



from pathlib import PurePath
from typing import Dict, Optional, List, Union

from labml import logger
from labml.internal import util
from labml.internal.configs.base import Configs
from labml.internal.configs.utils import Value
from labml.logger import Text

from .calculator import Calculator
from .parser import Parser

_CONFIG_PRINT_LEN = 40


class ConfigProcessorDict:
    def __init__(self, configs: Dict[str, any], values: Dict[str, any] = None):
        self.configs = configs
        if values is None:
            values = {}
        self.values = values

    def __call__(self, run_order: Optional[List[Union[List[str], str]]] = None):
        self.configs.update(self.values)

    def to_json(self):
        orders = {k: i for i, k in enumerate(self.configs.keys())}
        configs = {}
        for k, v in self.configs.items():
            configs[k] = {
                'name': k,
                'type': str(type(v)),
                'value': Value.to_yaml_truncated(v),
                'order': orders.get(k, -1),
                'options': [],
                'computed': Value.to_yaml_truncated(v),
                'is_hyperparam': False,
                'is_meta': False,
                'is_explicitly_specified': (k in self.values)
            }

        return configs

    def save(self, configs_path: PurePath):
        configs = self.to_json()

        with open(str(configs_path), "w") as file:
            file.write(util.yaml_dump(configs))

    def get_hyperparams(self):
        return self.values.copy()

    def __print_config(self, key, *, value=None):
        parts = ['\t']

        is_hyperparam = key in self.values

        if is_hyperparam:
            parts.append((key, [Text.key, Text.highlight]))
        else:
            parts.append((key, Text.key))

        parts.append((' = ', Text.subtle))

        value_str = Value.to_str(value)

        value_str = value_str.replace('\n', '')
        if len(value_str) < _CONFIG_PRINT_LEN:
            parts.append((f"{value_str}", Text.value))
        else:
            parts.append((f"{value_str[:_CONFIG_PRINT_LEN]}...", Text.value))
        parts.append('\t')

        return parts

    def print(self):
        order = list(self.configs.keys())
        order.sort()

        logger.log("Configs:", Text.heading)

        for k in order:
            parts = self.__print_config(k, value=self.configs[k])

            logger.log(parts)

        logger.log()



from typing import List, Dict, Type, Set, Optional, Union, Any, Tuple
from typing import TYPE_CHECKING

from .config_function import ConfigFunction
from .eval_function import EvalFunction
from .setup_function import SetupFunction
from ... import logger
from ... import monit

if TYPE_CHECKING:
    from .base import Configs
    from .processor import ConfigProcessor


class Calculator:
    secondary_values: Dict[str, Dict[str, any]]
    evals: Dict[str, Dict[str, EvalFunction]]
    options: Dict[str, Dict[str, ConfigFunction]]
    types: Dict[str, Type]
    values: Dict[str, any]

    configs: 'Configs'

    dependencies: Dict[str, Set[str]]
    topological_order: List[str]
    stack: List[str]
    visited: Set[str]
    is_computed: Set[str]
    is_top_sorted: Set[str]
    config_processors: Dict[str, 'ConfigProcessor']
    secondary_attributes: Dict[str, Set[str]]
    saved_configs: Dict[str, Any]

    def __init__(self, *,
                 configs: 'Configs',
                 options: Dict[str, Dict[str, ConfigFunction]],
                 evals: Dict[str, Dict[str, EvalFunction]],
                 types: Dict[str, Type],
                 values: Dict[str, any],
                 secondary_values: Dict[str, Dict[str, any]],
                 aggregate_parent: Dict[str, str]):
        self.secondary_values = secondary_values
        self.aggregate_parent = aggregate_parent
        self.evals = evals
        self.configs = configs
        self.options = options
        self.types = types
        self.values = values

        self.visited = set()
        self.stack = []
        self.is_top_sorted = set()
        self.topological_order = []
        self.is_computed = set()
        self.config_processors = {}
        self.secondary_attributes = {}
        self.saved_configs = {}
        self.cached = {}

    def __get_property(self, key: str) -> Tuple[Any, Optional[str], Optional[Union[ConfigFunction, SetupFunction]]]:
        if key in self.options:
            value = self.values[key]
            if value not in self.options[key]:
                return value, None, None
            return None, value, self.options[key][value]

        return self.values[key], None, None

    def __cache_calculated(self, key: str, option: Optional[str], value: any):
        if key not in self.cached:
            self.cached[key] = {}

        if option is not None:
            self.cached[key][option] = value

    def __get_dependencies(self, key) -> Set[str]:
        if key in self.options:
            value = self.values[key]
            if value not in self.options[key]:
                return set()
            sa = self.options[key][value].secondary_attributes
            for dep, att in sa.items():
                if dep not in self.secondary_attributes:
                    self.secondary_attributes[dep] = set()
                self.secondary_attributes[dep] = self.secondary_attributes[dep].union(att)
            return self.options[key][value].dependencies

        if key in self.evals:
            value = self.values.get(key, None)
            if not value:
                value = 'default'
            if value not in self.evals[key]:
                return set()
            return self.evals[key][value].dependencies

        return set()

    def __create_graph(self):
        self.dependencies = {}
        for k in self.types:
            self.dependencies[k] = self.__get_dependencies(k)
        for k in self.evals:
            self.dependencies[k] = self.__get_dependencies(k)

    def __check_missing(self, key: str):
        if key in self.values:
            return

        if key in self.evals:
            return

        raise RuntimeError(f"Cannot compute {key}, required by {self.user[key]}")

    def __add_to_topological_order(self, key):
        assert self.stack.pop() == key
        self.__check_missing(key)
        self.is_top_sorted.add(key)
        self.topological_order.append(key)

    def __traverse(self, key):
        for d in self.dependencies[key]:
            if d not in self.is_top_sorted:
                self.__add_to_stack(d, key)
                return

        self.__add_to_topological_order(key)

    def __add_to_stack(self, key: str, user: Optional[str]):
        if key in self.is_top_sorted:
            return

        if key in self.visited:
            raise RuntimeError(f"Cyclic dependency: {key}", self.stack)

        self.visited.add(key)
        self.stack.append(key)
        self.user[key] = user

    def __dfs(self):
        while len(self.stack) > 0:
            key = self.stack[-1]
            self.__traverse(key)

    def __topological_sort(self, keys: List[str]):
        for k in keys:
            assert k not in self.is_top_sorted

        for k in keys:
            self.__add_to_stack(k, None)
            self.__dfs()

    def __set_configs(self, key, value, is_direct: bool):
        assert key not in self.is_computed
        from .base import Configs
        from .processor import ConfigProcessor
        if isinstance(value, Configs):
            primary = value.__dict__.get('_primary', None)
            processor = ConfigProcessor(value,
                                        self.secondary_values.get(key, None),
                                        is_directly_specified=is_direct)
            secondary_attributes = self.secondary_attributes.get(key, set())
            if primary:
                secondary_attributes.add(primary)
            processor(list(secondary_attributes))
            if primary:
                value = value.__dict__.get(primary)
            self.config_processors[key] = processor

        self.is_computed.add(key)
        self.configs.__setattr__(key, value)
        self.saved_configs[key] = value

    def __compute(self, key):
        if key in self.is_computed:
            return

        if key in self.evals:
            return

        value, option, func = self.__get_property(key)

        if option is None:
            self.__set_configs(key, value, is_direct=True)
        elif key in self.cached and option in self.cached[key]:
            self.__set_configs(key, self.cached[key][option], is_direct=False)
        elif isinstance(func, ConfigFunction):
            s = monit.section(f'Prepare {key}', is_new_line=False)
            with s:
                value = func(self.configs)
            if s.get_estimated_time() >= 0.01:
                logger.log()
            else:
                logger.log(' ' * 100, is_new_line=False)

            if type(func.config_names) == str:
                self.__set_configs(key, value, is_direct=False)
            else:
                if not isinstance(value, tuple) and not isinstance(value, list):
                    raise RuntimeError(f"Expect a tuple or a list as results for {func.config_names}")
                if len(value) != len(func.config_names):
                    raise RuntimeError(f"Number of items in results {func.config_names}"
                                       f" for should match the number of configs")

                for i, k in enumerate(func.config_names):
                    if k == key:
                        self.__set_configs(k, value[i], is_direct=False)
                    else:
                        self.__cache_calculated(k, option, value[i])
        elif isinstance(func, SetupFunction):
            s = monit.section(f'Prepare {key}', is_new_line=False)
            with s:
                func(self.configs)
            if s.get_estimated_time() >= 0.01:
                logger.log()
            else:
                logger.log(' ' * 100, is_new_line=False)

            for i, k in enumerate(func.config_names):
                if k == key:
                    self.__set_configs(k, self.configs.__getattribute__(k), is_direct=False)
                elif k in self.is_computed:
                    self.configs.__setattr__(k, self.saved_configs[k])
                else:
                    self.__cache_calculated(k, option, self.configs.__getattribute__(k))
        else:
            raise RuntimeError(f"Unknown calculator function for {key}", func)

    def __compute_values(self):
        for k in self.topological_order:
            if k not in self.is_computed:
                self.__compute(k)

        parents = set()
        for k in self.topological_order:
            if k in self.aggregate_parent:
                parent = self.aggregate_parent[k]
                if parent not in self.is_computed:
                    parents.add(parent)

        self.topological_order = list(parents) + self.topological_order

    def __call__(self, run_order: Optional[List[Union[List[str], str]]]):
        if run_order is None:
            run_order = [list(self.types.keys())]
        run_order: List[Union[List[str], str]]

        for i in range(len(run_order)):
            keys = run_order[i]
            if type(keys) == str:
                run_order[i] = [keys]

        s = monit.section('Calculate config dependencies', is_new_line=False)
        with s:
            self.__create_graph()
        if s.get_estimated_time() >= 0.01:
            logger.log()

        self.visited = set()
        self.user = {}
        self.stack = []
        self.is_top_sorted = set()
        self.topological_order = []
        self.is_computed = set()

        for keys in run_order:
            self.__topological_sort(keys)
            self.__compute_values()
